// COPYRIGHT 2008 ARTEM KUCHIN
// www.artem.ru
// artem@artem.ru
// You may use this code however you want, but just give me a credit and
// leave a link to my site.


#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>
#include <avr/eeprom.h> 
#include <avr/pgmspace.h>
#include "usbdrv.h"


// averaging cycles
#define AVGN 10 // DO NOT SET HIGHER THAN 15 if timer prescaler is 256!

// SENS defines a threshold for "button stepped on" value
// for example 5 means 1/5 more than in unused state, that is 20% more.
// 5 prooved to be a good value

//                           R  D  L  U  UL UR
static unsigned int SENS[6]={10,06,10,07,10,10};

// hysteresis, 20 means 1/20 of unpressed values.
static unsigned int HYST[6]={30,20,30,30,20,20};

// example:
// unpress value is 1000 (just an example)
// SENS 15, AVGN 30, HYST 20
// then 
//    the value measured for 30 cycles
//    if the result is more than 1000+1000/15 then the button is considered pressed
// after the button is considered pressed  
//    if the values is less than 1000+1000/20 then button is unpressed again


typedef struct
{
  uchar x; //Byte-0, X (0, 1, 2)
  uchar y; //Byte-1, Y (0, 1, 2)
  union {
    uchar buttons; //Byte-2, buttons 4
    struct
    {
      uchar btn1: 1; //0, 1
      uchar btn2: 1; //0, 1
      uchar btn3: 1; //0, 1
      uchar btn4: 1; //0, 1
      uchar btn5: 1; //0, 1
      uchar btn6: 1; //0, 1
      uchar btn7: 1; //0, 1
      uchar btn8: 1; //0, 1     
    } b;
  } u;
} t_Controller;

t_Controller data[2];


//=============================================================================
//USB report
static uchar usbReport[1+sizeof(t_Controller)]; //First byte of ReportID
//=============================================================================
//HID
PROGMEM char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] =
{
  0x05, 0x01, // USAGE_PAGE (Generic Desktop)
  0x09, 0x04, // USAGE (Joystick)
  0xa1, 0x01, // COLLECTION (Application)
  0x09, 0x01, //   USAGE (Pointer)
  0xa1, 0x00, //   COLLECTION (Physical)
  0x85, 0x01, //     REPORT_ID (1)
  
  0x09, 0x30, //     USAGE (X)
  0x09, 0x31, //     USAGE (Y)
  0x15, 0, //     LOGICAL_MINIMUM (0)
  0x25, 2, //     LOGICAL_MAXIMUM (2)
  0x75, 0x08, //     REPORT_SIZE (8)
  0x95, 0x02, //     REPORT_COUNT (2)
  0x81, 0x02, //     INPUT (Data,Var,Abs)

  0x15, 0x00, //     LOGICAL_MINIMUM (0)
  0x25, 0x01, //     LOGICAL_MAXIMUM (1)
  0x95, 0x06, //     REPORT_COUNT (6)
  0x75, 0x01, //     REPORT_SIZE (1)
  0x05, 0x09, //     USAGE_PAGE (Button)
  0x19, 0x01, //     USAGE_MINIMUM (Button 1)
  0x29, 0x06, //     USAGE_MAXIMUM (Button 6)
  0x81, 0x02, //     INPUT (Data,Var,Abs)

  0x95, 0x02, //     REPORT_COUNT (2)
  0x75, 0x01, //     REPORT_SIZE (1)
  0x81, 0x01, //     INPUT (Constant) - 2 bit padding   

  0xc0,       //   END_COLLECTION
  0xc0,       // END_COLLECTION
};
//-----------------------------------------------------------------------------
//Output ReportID
static uchar *usbBuildReport(uchar id)
{
  static uchar lastReport = 0; //Last ReportID

  if (id) lastReport = id; //Real ReportID
  else lastReport = (lastReport %4) + 1; //Next ReportID

  //Copy report to buffer
  uchar *dp = (void *) &data[lastReport-1];
  uchar *rp = usbReport;

  //Fill ReportID
   *rp++ = lastReport;
  for (unsigned char i = 0; i < sizeof(t_Controller); i++) *rp++ = *dp++;

  //Return with address of buffer
  return usbReport;
}
//-----------------------------------------------------------------------------
//Fuction from Hidkeys
uchar usbFunctionSetup(uchar data[8])
{
  usbRequest_t *rq = (void *)data;

  if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
  {
    // Return the report if host requests it without USB interrupt
    if (rq->bRequest == USBRQ_HID_GET_REPORT)
    {
      // wValue: ReportType (highbyte), ReportID (lowbyte)
      // build report requested
      usbMsgPtr = usbBuildReport(rq->wValue.bytes[0]);
      return sizeof(usbReport);
    }
  }
  return 0;
}


/* ------------------------------------------------------------------------- */

int main(void)
{
	int cnt=0;
	int tt=0;
	char s=0; 
	char a=0;
	uint8_t tv;
	uint8_t b=0;
	uint8_t i;
	int cc[6];
	char down[6]={0,0,0,0,0,0};
	uint16_t th[6]= {0,0,0,0,0,0};
	uint16_t thh[6]= {0,0,0,0,0,0}; // hysteresis threshold
	uint8_t t=1;
	uint8_t mode; // 0 - only buttons, 1 - two buttons [0 and 1] and two axis (X,Y)
	uint8_t button_state=0;
	uint16_t bcnt=0;

	// array index for leds and pads (only 4 leds)
	// 0 - right
	// 1 - dopn
	// 2 - left
	// 3 - up
	// 4 - up-left
	// 5 - up-right

	// leds 
	volatile uint8_t *led_ddr[4] ={&DDRB,&DDRB,&DDRB,&DDRB};
	volatile uint8_t *led_port[4]={&PORTB,&PORTB,&PORTB,&PORTB};
	volatile uint8_t  led_bit[4] ={_BV(5),_BV(4),_BV(3),_BV(2)};

    // pad buttons
    volatile uint8_t *pad_ddr_out[6]   ={&DDRD ,&DDRC ,&DDRC ,&DDRC ,&DDRB ,&DDRD};
	volatile uint8_t *pad_port_out[6]  ={&PORTD,&PORTC,&PORTC,&PORTC,&PORTB,&PORTD};
	volatile uint8_t pad_bit_out[6]    ={_BV(1),_BV(3),_BV(1),_BV(0),_BV(0),_BV(6)};
	
	volatile uint8_t *pad_ddr_in[6]    ={&DDRD ,&DDRC ,&DDRC ,&DDRB ,&DDRD ,&DDRD};
	volatile uint8_t *pad_port_in[6]   ={&PORTD,&PORTC,&PORTC,&PORTB,&PORTD,&PORTD};
	volatile uint8_t *pad_pin_in[6]    ={&PIND ,&PINC ,&PINC ,&PINB ,&PIND ,&PIND};
	volatile uint8_t pad_bit_in[6]     ={_BV(0),_BV(4),_BV(2),_BV(1),_BV(7),_BV(5)};

	unsigned int calibration=1;


	eeprom_write_word ((uint16_t*)64, 0x1234);
	// read last saved mode
	mode=eeprom_read_byte ((uint8_t *)48);
	if (mode==0xff) mode=0;

	// capacitor sensor setup
	for(i=0;i<6;i++) {
		// turn on output pin and set it low
		*pad_ddr_out[i]  |= pad_bit_out[i];
		*pad_port_out[i] &= ~pad_bit_out[i];

		// turn on input pin and turn off pull up resistor
		*pad_ddr_in[i]  &= ~pad_bit_in[i];
		*pad_port_in[i] &= ~pad_bit_in[i];
	}

    //setup leds ports to led output mode
	for(i=0;i<4;i++) {
		*led_ddr[i] |= led_bit[i];
		*led_port[i] &= ~led_bit[i];
	}


	// button - PC5
	DDRC &= ~_BV(5);  // input
	PORTC |= _BV(5); // pull up on


	// setup timer with good speed
	TCNT0=0;
	// prescaler devisor is 1024 (16MHZ/1024=15625 ticks per second, 15625/256=61 oveflows per second)
	// prescaler devisor is 256 (162500 ticks per second, 15625/256=244 oveflows per second)
	TCCR0=_BV(CS02); // prescaler is 256



	for(;;){
		// var init
		cnt=0;
		s=0; 
		b=0;
		t=0;
		button_state=0;
		for(i=0;i<6;i++){
			down[i]=0;
			th[i]=0;
		}
		calibration=1;

		// init joystick data
		data[0].x=1;
		data[0].y=1;
		data[0].u.b.btn1=0;
		data[0].u.b.btn2=0;
		data[0].u.b.btn3=0;
		data[0].u.b.btn4=0;
		data[0].u.b.btn5=0;
		data[0].u.b.btn6=0;

    	usbInit();
    	usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
		_delay_ms(500);
    	usbDeviceConnect();
    	sei();

    	for(;;){                /* main event loop */
        	usbPoll();
        	if(usbInterruptIsReady()){
            	/* called after every poll of the interrupt endpoint */
				if(!calibration){
					if (mode){
						data[0].u.b.btn1=down[4];
						data[0].u.b.btn2=down[5];
						data[0].x=down[0]-down[2]+1;
						data[0].y=down[1]-down[3]+1;
					}
					else {
						data[0].u.b.btn1=down[0];
						data[0].u.b.btn2=down[1];
						data[0].u.b.btn3=down[2];
						data[0].u.b.btn4=down[3];
						data[0].u.b.btn5=down[4];
						data[0].u.b.btn6=down[5];
					}
				}
				usbSetInterrupt(usbBuildReport(0), sizeof(usbReport));
        	}

			switch(button_state){
				case 0:
					// button pressed
					if(!(PINC & _BV(5))){
						button_state=1;
						bcnt=0;
					}
					break;
				case 1:
					// anti jitter pause
					bcnt++;
					if (bcnt>1000){
						button_state=2;
					}
					break;
				case 2:
					// button released
					if(PINC & _BV(5)){
						button_state=3;
						mode^=1; // switch mode
						eeprom_write_byte ((uint8_t *)48, mode);
					}
					break;
					
			}
			if (button_state==3){
				button_state=0;
				break; // reinit device
			}
		
			switch (s){
				case 0: 
					// start measure  and averaging
					tt=0;
					cnt=0;
					a=0;
					// reset timer
					TCNT0=0;
					s=1;
					break;
				case 1:
					// start discharge of all Pads
					for(i=0;i<6;i++) {
						*pad_port_out[i] &= ~pad_bit_out[i];
					}
					s=2;
					break;
				case 2:
					if( !( *pad_pin_in[b] & pad_bit_in[b])){
						s=3;
					}
					break;
				case 3:
					// start charge
					*pad_port_out[b] |= pad_bit_out[b];
					s=4;
					break;
				case 4:
					// check charge and count
					if (*pad_pin_in[b] & pad_bit_in[b]){
						// charge finished
						tv=TCNT0;
						a++;
						if (a>AVGN){
							cc[b]=((int)tv+cnt*256);
							if(calibration){
								calibration++;

								if (calibration>2400){
									if (cc[b]>th[b]){
										th[b]=cc[b];
									}
									if (calibration % 200 == 0){
										if(mode){
											if (t){
												*led_port[0] ^= led_bit[0];
												*led_port[2] ^= led_bit[2];
											}
											else {
												*led_port[1] ^= led_bit[1];
												*led_port[3] ^= led_bit[3];
											}
											t++;
											if (t>1)t=0;

										}
										else {
											*led_port[t] ^= led_bit[t];
											t++;
											if(t>3)t=0;
										}
									}
								}
							
								if (calibration>12000){
									calibration=0;
									for (i=0;i<6;i++){
										thh[i]=th[i]+th[i]/HYST[i];
										th[i]+=th[i]/SENS[i];
										eeprom_write_word ((uint16_t*)(i*2), th[i]);
									}					
									for (i=0;i<4;i++){		
										*led_port[i] &= ~led_bit[i];
									}
								}
							}
							else {
								// real work, calibaration finished
								if (down[b]){
									if(cc[b]<thh[b]){
										down[b]=0;
										if (b<4){
											*led_port[b] &= ~led_bit[b];
										}
									}
								}
								else {
									if(cc[b]>th[b]) {
										down[b]=1;
										if (b<4){
											*led_port[b] |= led_bit[b];
										}
									}
								}
								//
								//else {
								//	down[b]=0;
								//	if (b<4){
								//		*led_port[b] &= ~led_bit[b];
								//	}
								//}
							}
							b++;
							if (b>5){b=0;}
							s=0;
						}
						else {
							s=1;
						}
					}
					break;
			}

			if (TIFR & _BV(TOV0)){
				// over flow
				// reset overflow flag, it is not reset if no interrupt generated
				TIFR|=_BV(TOV0); 
				cnt++;
			}		
    	}
	}
}


